#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

import GridMap as gmap
import GridPlanner as gplanner

from pathlib import Path

import time
import random

import pickle

PURSUER = 1
EVADER = 2

GREEDY = "GREEDY"
MONTE_CARLO = "MONTE_CARLO"
VALUE_ITERATION = "VALUE_ITERATION"


class Player:
    def __init__(self, robots, role, policy=GREEDY, color='r', timeout=5.0, game=None, distance_matrix=None, epsilon=1):
        """
        Parameters
        ----------
        robots: list((in,int))
            coordinates of individual player's robots
        role: int
            player's role in the game
        """
        # list of the player's robots
        self.robots = robots[:]
        # next position of the player's robots
        self.next_robots = robots[:]

        if role == "EVADER":
            self.role = EVADER
        elif role == "PURSUER":
            self.role = PURSUER
        else:
            raise ValueError('Unknown player role')

        # selection of the policy
        if policy == GREEDY:
            self.policy = self.greedy_policy
        elif policy == MONTE_CARLO:
            self.policy = self.monte_carlo_policy
        elif policy == VALUE_ITERATION:
            self.policy = self.value_iteration_policy
        else:
            raise ValueError('Unknown policy')

        self.color = color  # color for plotting purposes
        self.timeout = timeout  # planning timeout fo MCTS
        self.game_name = game  # game name for loading value iteration policies

        # values for the value iteration policy
        self.values = None

        # instantiation of helper planner
        self.Planner = gplanner.GridPlanner()
        
        self.distance_matrix = distance_matrix

    #####################################################
    # Game interface functions
    #####################################################
    def load_distance_matrix(self):
        with open(self.game_name + ".dist", 'rb') as file:
            return pickle.load(file)
    
    def add_robot(self, pos):
        """
        method to add a robot to the player
        
        Parameters
        ----------
        pos: (int,int)
            position of the robot
        """
        self.robots.append(pos)
        self.next_robots.append(pos)

    def del_robot(self, pos):
        """
        method to remove the player's robot 

        Parameters
        ----------
        pos: (int,int)
            position of the robot to be removed
        """
        idx = self.robots.index(pos)
        self.robots.pop(idx)
        self.next_robots.pop(idx)

    def calculate_step(self, gridmap, evaders, pursuers):
        """
        method to calculate the player's next step using selected policy
        
        Parameters
        ----------
        gridmap: GridMap
            map of the environment
        evaders: list((int,int))
            list of coordinates of evaders in the game (except the player's robots, if he is evader)
        pursuers: list((int,int))
            list of coordinates of pursuers in the game (except the player's robots, if he is pursuer)
        """        
        self.policy(gridmap, evaders, pursuers)

    def floyd_warshall(self, gridmap):
        self.distance_matrix = {}
        passable = []
        for i in range(gridmap.width):
            for j in range(gridmap.height):
                if gridmap.passable((i, j)):
                    passable.append((i, j))
                    for i_new in range(gridmap.width):
                        for j_new in range(gridmap.height):
                            if gridmap.passable((i_new, j_new)):
                                if i == i_new and j == j_new:
                                    self.distance_matrix[((i, j), (i_new, j_new))] = 0
                                elif (i, j) in gridmap.neighbors4((i_new, j_new)):
                                    self.distance_matrix[((i, j), (i_new, j_new))] = 1
                                else:
                                    self.distance_matrix[((i, j), (i_new, j_new))] = np.inf
        print(passable)
        for k in passable:
            for i in passable:
                for j in passable:
                    self.distance_matrix[(i, j)] = min(self.distance_matrix[(i, j)],
                                                       self.distance_matrix[(i, k)] + self.distance_matrix[(k, j)])
        with open(str(len(passable)) + ".dist", "wb") as file:
            pickle.dump(self.distance_matrix, file)

    def take_step(self):
        """
        method to perform the step 
        """
        self.robots = self.next_robots[:]

    #####################################################
    #####################################################
    # GREEDY POLICY
    #####################################################
    #####################################################
    def greedy_policy(self, gridmap, evaders, pursuers, epsilon=1):
        """
        Method to calculate the greedy policy action
        
        Parameters
        ----------
        gridmap: GridMap
            Map of the environment
        evaders: list((int,int))
            list of coordinates of evaders in the game (except the player's robots, if he is evader)
        pursuers: list((int,int))
            list of coordinates of pursuers in the game (except the player's robots, if he is pursuer)
        epsilon: float (optional)
            optional epsilon-greedy parameter
        """
        self.next_robots = self.robots[:]

        # for each of player's robots plan their actions
        for idx in range(0, len(self.robots)):
            robot = self.robots[idx]

            # TODO: implement greedy policy

            ##################################################
            # RANDOM Policy
            ##################################################
            # extract possible coordinates to go (actions)
            neighbors = gridmap.neighbors4(robot)
            # introducing randomness in neighbor selection
            random.shuffle(neighbors)

            # select random goal
            pos_selected = neighbors[0]
            ##################################################

            # select the next action based on own role
            if self.role == PURSUER:
                self.next_robots[idx] = pos_selected

            if self.role == EVADER:
                self.next_robots[idx] = pos_selected

    #####################################################
    #####################################################
    # MONTE CARLO TREE SEARCH POLICY
    #####################################################
    #####################################################
    def monte_carlo_policy(self, gridmap, evaders, pursuers):
        """
        Method to calculate the monte carlo tree search policy action
        
        Parameters
        ----------
        gridmap: GridMap
            Map of the environment
        evaders: list((int,int))
            list of coordinates of evaders in the game (except the player's robots, if he is evader)
        pursuers: list((int,int))
            list of coordinates of pursuers in the game (except the player's robots, if he is pursuer)
        """
        self.next_robots = self.robots[:]

        # for each of player's robots plan their actions
        for idx in range(0, len(self.robots)):
            robot = self.robots[idx]

            # measure the time for selecting next action
            clk = time.time()
            
            # TODO: implement MCTS policy

            ##################################################
            # RANDOM Policy
            ##################################################
            # extract possible coordinates to go (actions)
            neighbors = gridmap.neighbors4(robot)
            # introducing randomness in neighbor selection
            random.shuffle(neighbors)

            # select random goal
            pos_selected = neighbors[0]
            ##################################################

            # select the next action
            self.next_robots[idx] = pos_selected

    #####################################################
    #####################################################
    # VALUE ITERATION POLICY
    #####################################################
    #####################################################
    def value_iteration_policy(self, gridmap, evaders, pursuers):
        """
        Method to calculate the value-iteration policy action
        
        Parameters
        ----------
        gridmap: GridMap
            Map of the environment
        evaders: list((int,int))
            list of coordinates of evaders in the game (except the player's robots, if he is evader)
        pursuers: list((int,int))
            list of coordinates of pursuers in the game (except the player's robots, if he is pursuer)
        """
        self.next_robots = self.robots[:]

        # if there are not precalculated values for policy
        if not self.values:
            policy_file = Path("policies/" + self.game_name + ".policy")
            ################################################### 
            # if there is policy file, load it...
            ################################################### 
            if policy_file.is_file():
                # load the strategy file
                self.values = pickle.load(open(policy_file, 'rb'))

                ###################################################
            # ...else calculate the policy
            ################################################### 
            else:
                # TODO:calculate the Value iteration policy
                self.values = None

                # save the policy
                pickle.dump([self.values], open(policy_file, 'wb'))

        for idx in range(0, len(self.robots)):
            robot = self.robots[idx]

            #############################################
            # TODO: decide the next action  based on the policy
            #############################################
            neighbors = gridmap.neighbors4(robot)
            # introducing randomness in neighbor selection
            random.shuffle(neighbors)

            # select random goal
            pos_selected = neighbors[0]
            #############################################

            # select the next action for the robots
            self.next_robots[idx] = pos_selected

    #####################################################
    # Helper functions
    #####################################################
    def dist(self, gridmap, coord1, coord2):
        return self.distance_matrix[(coord1, coord2)]
